﻿//---------------------------------------------------------------------------
//
//   Single Move
//
//          Copyright(C) 2007-2012, tetraface Inc.
//
//     Sample for Command plug-in for moving a single vertex.
//     Created DLL must be installed in "Plugins\Command" directory.
//    After installing the plug-in, the "Single Move" button will be 
//    shown in the command panel.
//
//    　単一の頂点を移動するコマンドプラグインのサンプル。
//    　作成したDLLは"Plugins\Command"フォルダに入れる必要がある。
//    　プラグインをインストールすると"Single Move"ボタンが
//    コマンドパネルに追加されます。
//
//
//---------------------------------------------------------------------------
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <float.h>
#include "MQBasePlugin.h"
#include "resource.h"

HINSTANCE g_hInstance;


class SingleMovePlugin : public MQCommandPlugin
{
public:
	// コンストラクタ
	SingleMovePlugin();

	// プラグインIDを返す。
	virtual void GetPlugInID(DWORD *Product, DWORD *ID);
	// プラグイン名を返す。
	virtual const char *GetPlugInName(void);
	// ポップアップメニューに表示される文字列を返す。
	virtual const char *EnumString(void);
	// アプリケーションの初期化
	virtual BOOL Initialize();
	// アプリケーションの終了
	virtual void Exit();
	// 表示・非表示切り替え要求
	virtual BOOL Activate(MQDocument doc, BOOL flag);
	// 描画時の処理
	virtual void OnDraw(MQDocument doc, MQScene scene, int width, int height);

	// 左ボタンが押されたとき
	virtual BOOL OnLeftButtonDown(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state);
	// 左ボタンが押されながらマウスが移動したとき
	virtual BOOL OnLeftButtonMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state);
	// 左ボタンが離されたとき
	virtual BOOL OnLeftButtonUp(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state);
	// マウスが移動したとき
	virtual BOOL OnMouseMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state);

private:
	int m_iHighlightObject;
	int m_iHighlightVertex;
	int m_iMoveObject;
	int m_iMoveVertex;
	bool m_bMoved;
	MQPoint m_ScreenPos;
	MQPoint m_MovedPos;

	HCURSOR m_MoveCursor;
};



//---------------------------------------------------------------------------
//  SingleMovePlugin::SingleMovePlugin
//    コンストラクタ
//---------------------------------------------------------------------------
SingleMovePlugin::SingleMovePlugin()
{
	m_iHighlightObject = -1;
	m_iHighlightVertex = -1;
	m_iMoveObject = -1;
	m_iMoveVertex = -1;
	m_bMoved = false;

	m_MoveCursor = NULL;
}


//---------------------------------------------------------------------------
//  SingleMovePlugin::GetPlugInID
//    プラグインIDを返す。
//---------------------------------------------------------------------------
void SingleMovePlugin::GetPlugInID(DWORD *Product, DWORD *ID)
{
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0x3901EA9B;
}


//---------------------------------------------------------------------------
//  SingleMovePlugin::GetPlugInName
//    プラグイン名を返す。
//---------------------------------------------------------------------------
const char *SingleMovePlugin::GetPlugInName(void)
{
	return "Single Move       Copyright(C) 2007-2012, tetraface Inc.";
}


//---------------------------------------------------------------------------
//  SingleMovePlugin::EnumString
//    ボタンに表示される文字列を返す。
//---------------------------------------------------------------------------
const char *SingleMovePlugin::EnumString(void)
{
	return "Single Move";
}


//---------------------------------------------------------------------------
//  SingleMovePlugin::Initialize
//    アプリケーションの初期化
//---------------------------------------------------------------------------
BOOL SingleMovePlugin::Initialize()
{
	// Load a cursor.
	// カーソルをロード
	if(m_MoveCursor == NULL){
		m_MoveCursor = LoadCursor(g_hInstance, MAKEINTRESOURCE(IDC_CURSOR1));
	}

	return TRUE;
}

//---------------------------------------------------------------------------
//  SingleMovePlugin::Exit
//    アプリケーションの終了
//---------------------------------------------------------------------------
void SingleMovePlugin::Exit()
{
	// Nothing to do.
	// 特に何もしない
}

//---------------------------------------------------------------------------
//  SingleMovePlugin::Activate
//    表示・非表示切り替え要求
//---------------------------------------------------------------------------
BOOL SingleMovePlugin::Activate(MQDocument doc, BOOL flag)
{
	if (!flag){
		// Erase a highlight
		// ハイライト表示を消す
		if (m_iHighlightObject != -1 && m_iHighlightVertex != -1){
			m_iHighlightObject = -1;
			m_iHighlightVertex = -1;
			RedrawAllScene();
		}
	}
	
	// It returns 'flag' as it is.
	// そのままflagを返す
	return flag;
}


//---------------------------------------------------------------------------
//  SingleMovePlugin::OnDraw
//    描画時の処理
//---------------------------------------------------------------------------
void SingleMovePlugin::OnDraw(MQDocument doc, MQScene scene, int width, int height)
{
	// Highlight the vertex.
	// 頂点をハイライト描画
	if (m_iHighlightObject != -1 && m_iHighlightVertex != -1)
	{
		MQObject obj = doc->GetObject(m_iHighlightObject);
		if (obj == NULL) return; // あり得ないはずだが、念のためNULLチェック

		// Calculate the position of the vertex.
		// 頂点位置を計算
		MQPoint sp = scene->Convert3DToScreen(obj->GetVertex(m_iHighlightVertex));
		MQPoint newvp = scene->ConvertScreenTo3D(MQPoint(sp.x,sp.y,0.00001f)); // 手前に描画するためZにほぼ0（計算誤差を防ぐため）をセット

		// Create a drawing object.
		// 描画オブジェクトを生成
		int vertex[1];
		MQObject draw = CreateDrawingObject(doc, DRAW_OBJECT_POINT);
		draw->SetColor(MQColor(1,1,0));
		draw->SetColorValid(TRUE);
		vertex[0] = draw->AddVertex(newvp);
		draw->AddFace(1, vertex);
	}
}


//---------------------------------------------------------------------------
//  SingleMovePlugin::OnLeftButtonDown
//    左ボタンが押されたとき
//---------------------------------------------------------------------------
BOOL SingleMovePlugin::OnLeftButtonDown(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	MQPoint clickpos(state.MousePos.x, state.MousePos.y, 0);
	float mindis2 = 10*10; //10ピクセル以内

	// It gets edit options.
	// 編集オプションを取得
	EDIT_OPTION option;
	GetEditOption(option);

	m_iMoveObject = -1;
	m_iMoveVertex = -1;
	m_bMoved = false;

	// Find a nearest vertex from the clicked position.
	// クリックされた位置に最も近い頂点を見つける
	int numObj = doc->GetObjectCount();
	for (int oi=0; oi<numObj; oi++)
	{
		MQObject obj = doc->GetObject(oi);
		if (obj == NULL) continue;

		// An invisible or locked object is not moved.
		// 非表示またはロックされていれば対象外
		if (!obj->GetVisible() || obj->GetLocking()) continue;

		// It edits a current object only.
		// カレントオブジェクトのみ編集
		if (option.CurrentObjectOnly && oi != doc->GetCurrentObjectIndex())
			continue;

		int numVertex = obj->GetVertexCount();
		for (int vi=0; vi<numVertex; vi++)
		{
			if (obj->GetVertexRefCount(vi) != 0)
			{
				// Is the vertex closed to the clicked position?
				// クリックされた位置に近いかどうか
				MQPoint sp = scene->Convert3DToScreen(obj->GetVertex(vi));
				float dis2 = (sp.x-clickpos.x)*(sp.x-clickpos.x) + (sp.y-clickpos.y)*(sp.y-clickpos.y);
				if (mindis2 > dis2){
					// It will move this vertex.
					// この頂点を移動対象に
					mindis2 = dis2;
					m_iMoveObject = oi;
					m_iMoveVertex = vi;
					m_ScreenPos = sp;
				}
			}
		}
	}

	// It returns TRUE because an original action was done. A default action will be ignored.
	// 標準動作の代わりに独自処理を行ったのでTRUEを返す
	return TRUE;
}


//---------------------------------------------------------------------------
//  SingleMovePlugin::OnLeftButtonMove
//    左ボタンが押されながらマウスが移動したとき
//---------------------------------------------------------------------------
BOOL SingleMovePlugin::OnLeftButtonMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if (m_iMoveObject != -1 && m_iMoveVertex != -1)
	{
		MQObject obj = doc->GetObject(m_iMoveObject);
		if (obj == NULL) return TRUE; // あり得ないはずだが、念のためNULLチェック

		// 編集オプションを取得
		EDIT_OPTION option;
		GetEditOption(option);

		// 頂点を移動（スクリーン座標のZはそのままでX/Yのみ動かす）
		MQPoint vp = scene->ConvertScreenTo3D(MQPoint(state.MousePos.x, state.MousePos.y, m_ScreenPos.z));
		if (option.SnapGrid){
			// グリッドへの吸着
			vp = GetSnappedPos(scene, vp, option.SnapGrid);
		}
		obj->SetVertex(m_iMoveVertex, vp);

		// 再描画
		RedrawScene(scene);

		// 移動済みチェック
		m_bMoved = true;
		m_MovedPos = vp;
	}

	// It returns TRUE because an original action was done. A default action will be ignored.
	// 標準動作の代わりに独自処理を行ったのでTRUEを返す
	return TRUE;
}


//---------------------------------------------------------------------------
//  SingleMovePlugin::OnLeftButtonUp
//    左ボタンが離されたとき
//---------------------------------------------------------------------------
BOOL SingleMovePlugin::OnLeftButtonUp(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if (m_iMoveObject != -1 && m_iMoveVertex != -1 && m_bMoved)
	{
		// Redraw a scene and update an undo if a vertex was moved.
		// 頂点が移動された場合、再描画とアンドゥの更新を行う
		RedrawAllScene();
		UpdateUndo();

		// Send a message to StationSpy plug-in.
		// StationSpyプラグインにメッセージを送る
		char buf[128];
		sprintf_s(buf, sizeof(buf), "Moved to (%.3f %.3f %.3f)", m_MovedPos.x, m_MovedPos.y, m_MovedPos.z);
		int ret = SendUserMessage(doc, 0x56A31D20, 0x9CE001E3, "text", buf);
		sprintf_s(buf, sizeof(buf), "Received %d from SendUserMessage()\n", ret);
		OutputDebugString(buf);
	}

	// It returns TRUE because an original action was done. A default action will be ignored.
	// 標準動作の代わりに独自処理を行ったのでTRUEを返す
	return TRUE;
}


//---------------------------------------------------------------------------
//  SingleMovePlugin::OnMouseMove
//    マウスが移動したとき
//---------------------------------------------------------------------------
BOOL SingleMovePlugin::OnMouseMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	float mindis2 = 10*10; //10ピクセル以内

	int newoi = -1;
	int newvi = -1;

	// It gets edit options.
	// 編集オプションを取得
	EDIT_OPTION option;
	GetEditOption(option);

	// Find a nearest vertex from the clicked position.
	// クリックされた位置に最も近い頂点を見つける
	int numObj = doc->GetObjectCount();
	for (int oi=0; oi<numObj; oi++)
	{
		MQObject obj = doc->GetObject(oi);
		if (obj == NULL) continue;

		// An invisible or locked object is not moved.
		// 非表示またはロックされていれば対象外
		if (!obj->GetVisible() || obj->GetLocking()) continue;

		// It edits a current object only.
		// カレントオブジェクトのみ編集
		if (option.CurrentObjectOnly && oi != doc->GetCurrentObjectIndex())
			continue;

		int numVertex = obj->GetVertexCount();
		for (int vi=0; vi<numVertex; vi++)
		{
			if (obj->GetVertexRefCount(vi) != 0)
			{
				// Is the vertex closed to the clicked position?
				// クリックされた位置に近いかどうか
				MQPoint sp = scene->Convert3DToScreen(obj->GetVertex(vi));
				float dis2 = (sp.x-state.MousePos.x)*(sp.x-state.MousePos.x) + (sp.y-state.MousePos.y)*(sp.y-state.MousePos.y);
				if (mindis2 > dis2){
					// It will move this vertex.
					// この頂点を移動対象に
					mindis2 = dis2;
					newoi = oi;
					newvi = vi;
				}
			}
		}
	}

	// Redraw a scene when the cursor is on the different vertex from the previous time.
	// 前回と違う頂点上にカーソルがきた場合は再描画
	if (m_iHighlightObject != newoi || m_iHighlightVertex != newvi)
	{
		m_iHighlightObject = newoi;
		m_iHighlightVertex = newvi;

		if(m_iHighlightVertex != -1){
			SetMouseCursor(m_MoveCursor);
		}else{
			SetMouseCursor(GetResourceCursor(MQCURSOR_DEFAULT));
		}
		RedrawScene(scene);
	}

	// It returns FALSE for a default action.
	// 独自処理を行ったが、標準動作も行わせるためFALSEを返す
	return FALSE;
}



//---------------------------------------------------------------------------
//  GetPluginClass
//    プラグインのベースクラスを返す
//---------------------------------------------------------------------------
MQBasePlugin *GetPluginClass()
{
	static SingleMovePlugin plugin;
	return &plugin;
}



//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HINSTANCE hinstDLL, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	if(ul_reason_for_call == DLL_PROCESS_ATTACH){
		g_hInstance = hinstDLL;
	}
    return TRUE;
}

